#ifndef CORE_UTILITY_HPP
#define CORE_UTILITY_HPP

#include "core/Meta.hpp"
#include "core/Types.hpp"

#ifdef CHECK_MEMORY
void* operator new(size_t count, const char* file, int line);
void* operator new[](size_t count, const char* file, int line);
#endif

namespace Core {
    inline void nothing() {
    }

    template<typename T, typename C = int>
    C popCount(const T& t) {
        static constexpr C map[16] = {0, 1, 1, 2, 1, 2, 2, 3,
                                      1, 2, 2, 3, 2, 3, 3, 4};
        C sum = 0;
        for(size_t i = 0; i < sizeof(T) * 8; i += 4) {
            sum += map[(t >> i) & 0xF];
        }
        return sum;
    }

    using ExitHandler = void (*)(int, void*);
    [[noreturn]] void exitWithHandler(const char* file, int line, int value);
    void setExitHandler(ExitHandler h, void* data);
#define EXIT(exitValue) Core::exitWithHandler(__FILE__, __LINE__, exitValue)

    using OutOfMemoryHandler = void (*)(void*);
    void setOutOfMemoryHandler(OutOfMemoryHandler h, void* data);

#ifdef CHECK_MEMORY
    void* debugAllocateRaw(const char* file, int line, size_t n);
    void* debugZeroAllocateRaw(const char* file, int line, size_t n);
    void* debugReallocateRaw(const char* file, int line, void* p, size_t n);
    void debugDeallocateRaw(void* p);
    void printMemoryReport(void);
#define allocateRaw(n) debugAllocateRaw(__FILE__, __LINE__, n)
#define zeroAllocateRaw(n) debugZeroAllocateRaw(__FILE__, __LINE__, n)
#define reallocateRaw(p, n) debugReallocateRaw(__FILE__, __LINE__, p, n)
#define deallocateRaw(p) debugDeallocateRaw(p)
#define coreNew(type, ...) new(__FILE__, __LINE__) type(__VA_ARGS__)
#define coreNewN(type, n) new(__FILE__, __LINE__) type[n]
#define coreDelete(p) delete(p)
#define coreDeleteN(p) delete[](p)
#else
    void* allocateRaw(size_t n);
    void* zeroAllocateRaw(size_t n);
    void* reallocateRaw(void* p, size_t n);
    void deallocateRaw(void* p);
#define printMemoryReport() nothing()
#define coreNew(type, ...) new type(__VA_ARGS__)
#define coreNewN(type, n) new type[n]
#define coreDelete(p) delete(p)
#define coreDeleteN(p) delete[](p)
#endif

    bool sleepMillis(i64 millis);
    bool sleepNanos(i64 nanos);
    i64 getNanos(void);

    template<typename T>
    void bubbleSort(T* data, size_t n) {
        bool swapped = true;
        while(swapped && n > 0) {
            swapped = false;
            n--;
            for(size_t i = 0; i < n; i++) {
                if(data[i] > data[i + 1]) {
                    swap(data[i], data[i + 1]);
                    swapped = true;
                }
            }
        }
    }
}

#endif
